# Generated by Django 3.2.9 on 2021-12-04 02:51

import ckeditor.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("email_signals", "0003_auto_20211204_0211"),
    ]

    operations = [
        migrations.AlterField(
            model_name="signal",
            name="html_email",
            field=ckeditor.fields.RichTextField(
                blank=True, null=True, verbose_name="HTML email"
            ),
        ),
    ]



from django.test import SimpleTestCase
from django.core import mail
from .testcase import setup_settings
from .. import emailer


class TestSendMail(SimpleTestCase):
    """Unittests for the `send_mail` function in the `emailer` module."""

    def test_send_mail_no_template(self):
        """Tests that the function is able to send an email without a
        template."""
        emailer.send_mail(
            subject="Test Subject",
            plain_message="Test Message",
            html_message="<p>Test Message</p>",
            recipient_list=["test@test.com"],
            from_email="test@test.com",
        )
        self.assertEqual(len(mail.outbox), 1)

    def test_send_mail_with_template(self):
        """Tests that the function is able to send an email with a template."""

        setup_settings()

        emailer.send_mail(
            subject="Test Subject",
            plain_message="Test Message",
            html_message="<p>Test Message</p>",
            recipient_list=["test@test.com"],
            from_email="test@test.com",
            template="email_signals/tests/test_emailer.html",
        )
        self.assertEqual(len(mail.outbox), 1)



from django import forms
from django.template.loader import get_template
from django.template import TemplateDoesNotExist
from django.utils.html import format_html
from . import models
from .registry import registered_content_types
from .utils import get_param_from_obj
from .constraint_checker import comparison_requires_2_params


def render_js(cls):
    """Forcing all media forms to be deferred."""
    return [
        format_html(
            '<script defer src="{}"></script>', cls.absolute_path(path)
        )
        for path in cls._js
    ]


forms.widgets.Media.render_js = render_js


class SignalAdminForm(forms.ModelForm):
    class Meta:
        model = models.Signal
        fields = "__all__"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Limit the content type choices to the registered content types
        self.fields["content_type"].queryset = registered_content_types()

    def clean(self):
        cleaned_data = super().clean()
        if self.is_valid():
            self._clean_mailing_list()
        return cleaned_data

    def _clean_mailing_list(self):
        """The `mailing_list` field contains a string which corresponds to a
        method that should exist in the model. Check that the function does
        exists. OR The mailing list can be a list of emails, there should be at least one email.
        """
        mailing_list = self.cleaned_data["mailing_list"]
        content_type = self.cleaned_data["content_type"]

        if (
            not hasattr(content_type.model_class(), mailing_list)
            and "@" not in mailing_list
        ):
            raise forms.ValidationError(
                f"The model does not have a function called {mailing_list} and the mailing list does not contain a list of emails"
            )
        return mailing_list

    def clean_template(self):
        """Check that the template exists."""
        template = self.cleaned_data["template"]
        if not template:
            return template
        try:
            get_template(template)
        except TemplateDoesNotExist:
            raise forms.ValidationError(
                f"The template {template} does not exist"
            )
        return template


class SignalConstraintAdminForm(forms.ModelForm):
    class Meta:
        models = models.SignalConstraint
        fields = "__all__"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Set autocomplete off for all fields.
        for field in self.fields:
            self.fields[field].widget.attrs["autocomplete"] = "off"

    def clean(self):
        cleaned_data = super().clean()
        if self.is_valid():
            self._clean_comparison()
        return cleaned_data

    def clean_param_1(self):
        """Validate `param_1` can be found in either the signal `kwargs` or the
        model instance.
        """
        param_1 = self.cleaned_data["param_1"]

        # This isn't the best way to do this. But it's the only way I can think
        # accepting keys in the kwargs.
        # For now we'll accept `created` as this is a common parameter to check
        # on post creation.
        if param_1 == "created":
            return param_1

        valid, _ = get_param_from_obj(param_1, self.instance.signal.model)
        if not valid:
            raise forms.ValidationError(
                f"The model does not have a parameter called {param_1}"
            )
        return param_1

    def _clean_comparison(self):
        """Check that the comparison is valid for the given parameters.
        Depending on what the `comparison` is, it would limit the value for
        `param_1` and `param_2`.

        Note: Due to the order of fields, we need to preprend the underscore
        in the function name and call it manually as `param_2` is cleaned
        after 'comparison'.
        """

        comparison = self.cleaned_data["comparison"]
        param_2 = self.cleaned_data["param_2"]

        if comparison_requires_2_params(comparison) and not param_2:
            raise forms.ValidationError(
                "This comparison requires a second parameter"
            )

        if not comparison_requires_2_params(comparison) and param_2:
            raise forms.ValidationError(
                "This comparison does not require a second parameter"
            )

        return comparison



import typing as _t
import re


def exact(param_1: _t.Any, param_2: _t.Any) -> bool:
    """Check if two objects are equal."""
    return param_1 == param_2


def iexact(param_1: _t.Any, param_2: _t.Any) -> bool:
    """Check if two objects are equal, ignoring case."""
    if isinstance(param_1, str) and isinstance(param_2, str):
        return param_1.lower() == param_2.lower()
    return exact(param_1, param_2)


def contains(param_1: _t.Any, param_2: _t.Any) -> bool:
    """Check if param_1 contains param_2."""
    if isinstance(param_1, str) and isinstance(param_2, str):
        return param_2 in param_1
    return False


def icontains(param_1: _t.Any, param_2: _t.Any) -> bool:
    """Check if param_1 contains param_2, ignoring case."""
    if isinstance(param_1, str) and isinstance(param_2, str):
        return param_2.lower() in param_1.lower()
    return False


def gt(param_1: _t.Any, param_2: _t.Any) -> bool:
    """Check if param_1 is greater than param_2."""
    try:
        return float(param_1) > float(param_2)
    except (TypeError, ValueError):
        return False


def gte(param_1: _t.Any, param_2: _t.Any) -> bool:
    """Check if param_1 is greater than or equal to param_2."""
    try:
        return float(param_1) >= float(param_2)
    except (TypeError, ValueError):
        return False


def lt(param_1: _t.Any, param_2: _t.Any) -> bool:
    """Check if param_1 is less than param_2."""
    try:
        return float(param_1) < float(param_2)
    except (TypeError, ValueError):
        return False


def lte(param_1: _t.Any, param_2: _t.Any) -> bool:
    """Check if param_1 is less than or equal to param_2."""
    try:
        return float(param_1) <= float(param_2)
    except (TypeError, ValueError):
        return False


def startswith(param_1: _t.Any, param_2: _t.Any) -> bool:
    """Check if param_1 starts with param_2."""
    if isinstance(param_1, str) and isinstance(param_2, str):
        return param_1.startswith(param_2)
    return False


def istartswith(param_1: _t.Any, param_2: _t.Any) -> bool:
    """Check if param_1 starts with param_2, ignoring case."""
    if isinstance(param_1, str) and isinstance(param_2, str):
        return param_1.lower().startswith(param_2.lower())
    return False


def endswith(param_1: _t.Any, param_2: _t.Any) -> bool:
    """Check if param_1 ends with param_2."""
    if isinstance(param_1, str) and isinstance(param_2, str):
        return param_1.endswith(param_2)
    return False


def iendswith(param_1: _t.Any, param_2: _t.Any) -> bool:
    """Check if param_1 ends with param_2, ignoring case."""
    if isinstance(param_1, str) and isinstance(param_2, str):
        return param_1.lower().endswith(param_2.lower())
    return False


def regex(param_1: _t.Any, param_2: _t.Any) -> bool:
    """Check if param_1 matches param_2."""
    if isinstance(param_1, str) and isinstance(param_2, str):
        return re.match(param_2, param_1) is not None
    return False


def iregex(param_1: _t.Any, param_2: _t.Any) -> bool:
    """Check if param_1 matches param_2, ignoring case."""
    if isinstance(param_1, str) and isinstance(param_2, str):
        return re.match(param_2, param_1, re.IGNORECASE) is not None
    return False


def isnull(param_1: _t.Any, _=None) -> bool:
    """Check if param_1 is null."""
    return param_1 is None


def isnotnull(param_1: _t.Any, _=None) -> bool:
    """Check if param_1 is not null."""
    return param_1 is not None


def istrue(param_1: _t.Any, _=None) -> bool:
    """Check that param_1 is truthy."""
    return bool(param_1)


def isfalse(param_1: _t.Any, _=None) -> bool:
    """Check that param_1 is falsy."""
    return not bool(param_1)



from django.contrib import admin
from . import models, forms


class SignalConstraintInline(admin.TabularInline):
    form = forms.SignalConstraintAdminForm
    model = models.SignalConstraint
    extra = 0


@admin.register(models.Signal)
class SignalAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "content_type",
        "signal_type",
        "constraints_count",
        "active",
    )

    form = forms.SignalAdminForm
    change_form_template = "email_signals/admin/email_signals/signal/change_form.html"  # noqa: E501
    inlines = [SignalConstraintInline]

    class Media:
        css = {"all": ("email_signals/css/signal_change_form.min.css",)}
        js = ("email_signals/js/signal_change_form.min.js",)



import typing as _t
from django.db.models import Model
from .models import SignalConstraint
from . import constraint_methods, utils


class ConstraintChecker:
    """Determines whether a model instance satisfies a constraints and
    determines if it is able to raise a signal.
    """

    def __init__(
        self, instance: Model, constraints: _t.List[Model], signal_kwargs: dict
    ):
        """Initialise the checker with the instance and signal kwargs.

        Args:
            instance: The model instance on which a signal is to be raised.
            constraints: The constraints for the model instance
            signal_kwargs: The kwargs retrieved from the signal handler.
        """
        self.instance = instance
        self.signal_kwargs = signal_kwargs
        self.constraints = constraints

    def run_tests(self) -> bool:
        """Run all tests and return `True` if all tests pass."""
        for constraint in self.constraints:
            p1, p2 = self.get_params(constraint)
            if not self.check_constraint(p1, p2, constraint.comparison):
                return False
        return True

    def get_params(
        self, constraint: SignalConstraint
    ) -> _t.Tuple[_t.Any, _t.Any]:
        """Given a `constraint` param fields, retrieve actual values.

        Args:
            constraint: The constraint to retrieve the params for.

        Returns:
            A tuple of the actual values of the params.
        """

        return (
            self.get_param_1(constraint.param_1),
            self.get_param_2(constraint.param_2),
        )

    def get_param_1(self, param_1: str) -> _t.Any:
        """Given `param_1` from the constraints, retrieve it's actual value.

        Args:
            param_1: The param_1 field from the constraint.

        Returns:
            The actual value of the param_1 field.
        """
        if not param_1:
            raise ValueError("`param_1` is a required field.")

        success, param_1_val = utils.get_param_from_obj(
            param_1, self.signal_kwargs
        )
        if success:
            return param_1_val

        success, param_1_val = utils.get_param_from_obj(param_1, self.instance)
        if not success:
            raise ValueError(
                f"ContainsChecker: param_1 {param_1} not found in kwargs "
                "nor in model instance"
            )
        return param_1_val

    def get_param_2(self, param_2: str) -> _t.Any:
        """Given `param_2` from the constraints, retrieve it's actual value.

        Args:
            param_2: The param_2 field from the constraint.

        Returns:
            The actual value of the param_2 field.
        """
        if param_2 is None:
            return None

        success, param_2_val = utils.get_param_from_obj(
            param_2, self.signal_kwargs
        )
        if success:
            return param_2_val

        success, param_2_val = utils.get_param_from_obj(param_2, self.instance)
        if success:
            return param_2_val

        return utils.convert_to_primitive(param_2)

    @staticmethod
    def check_constraint(
        param_1: _t.Any, param_2: _t.Any, comparison: str
    ) -> bool:
        """Check if `param_1` and `param_2` satisfy the constraint.

        Args:
            param_1: First parameter to check.
            param_2: Second parameter to check.
            comparison: Comparison method to use.

        Returns:
            True if `param_1` and `param_2` satisfy the constraint.
        """

        method = getattr(constraint_methods, comparison, None)
        if method is None:
            raise ValueError(
                f"ContainsChecker: comparison {comparison} not found in "
                "constraint_methods"
            )
        return method(param_1, param_2)


def comparison_requires_2_params(comparison: str) -> bool:
    """Check if the comparison requires 2 params.

    Args:
        comparison: The comparison to check.

    Returns:
        True if the comparison requires 2 params.
    """
    return comparison in [
        "exact",
        "iexact",
        "contains",
        "icontains",
        "gt",
        "gte",
        "lt",
        "lte",
        "startswith",
        "istartswith",
        "endswith",
        "iendswith",
        "regex",
        "iregex",
    ]



import typing as _t
from django.contrib.contenttypes.models import ContentType
from .testcase import EmailSignalTestCase
from ..forms import SignalAdminForm
from ..registry import add_to_registry


class TestSignalAdminForm(EmailSignalTestCase):
    """Unitests for the `SignalAdminForm` class."""

    @classmethod
    def setUpClass(cls):
        """Creates a table for the model."""
        super().setUpClass()
        add_to_registry(cls.Customer)

    def sample_form(
        self, override_form_data: _t.Optional[dict] = None
    ) -> SignalAdminForm:
        form_data = {
            "name": "Test Signal",
            "description": "Test description",
            "content_type": ContentType.objects.get_for_model(
                self.customer_rec
            ).pk,
            "plain_message": "Test plain message",
            "html_message": "Test html message",
            "subject": "Test subject",
            "from_email": "test@email.com",
            "mailing_list": "my_mailing_list",
            "template": None,
            "signal_type": "pre_save",
        }
        form_data.update(override_form_data or {})
        form = SignalAdminForm(data=form_data)
        form.is_valid()
        return form

    def test_valid_form(self):
        """Basic test to check that a valid form is accepted."""
        form = self.sample_form()
        self.assertTrue(form.is_valid(), form.data)

    def test_invalid_mailing_list(self):
        """Test form where an invalid `mailing_list` is provided."""
        form = self.sample_form({"mailing_list": "invalid"})
        self.assertFalse(form.is_valid())

    def test_email_mailing_list(self):
        """Test form where an email list `mailing_list` is provided."""
        form = self.sample_form(
            {"mailing_list": "test@email.com,test1@email.com"}
        )
        self.assertTrue(form.is_valid())

    def test_invalid_template(self):
        """Test form where an invalid `template` is provided."""
        form = self.sample_form({"template": "invalid"})
        self.assertFalse(form.is_valid())

    def test_valid_template(self):
        """Test form where a valid `template` is provided."""
        form = self.sample_form(
            {"template": "email_signals/tests/test_emailer.html"}
        )
        self.assertTrue(form.is_valid(), form.data)



# Generated by Django 3.2.9 on 2021-12-01 22:06

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="Signal",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("description", models.TextField(blank=True, null=True)),
                ("plain_text_email", models.TextField(blank=True, null=True)),
                ("html_email", models.TextField(blank=True, null=True)),
                ("subject", models.CharField(max_length=255)),
                ("from_email", models.EmailField(max_length=254)),
                (
                    "to_emails_opt",
                    models.PositiveIntegerField(
                        default=1,
                        help_text="The choice of the user to send the email to. For each             integer `i`, the method `get_email_signal_emails_<i>` will be             called on the model instance. The method should return a list of             emails to send to.",
                    ),
                ),
                (
                    "template",
                    models.CharField(
                        blank=True,
                        help_text="Custom template to use for the email.",
                        max_length=100,
                        null=True,
                    ),
                ),
                (
                    "signal_type",
                    models.CharField(
                        choices=[
                            ("pre_save", "Pre Save"),
                            ("post_save", "Post Save"),
                            ("pre_delete", "Pre Delete"),
                            ("post_delete", "Post Delete"),
                        ],
                        max_length=20,
                    ),
                ),
                ("active", models.BooleanField(default=True)),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="SignalConstraint",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("param_1", models.CharField(max_length=255)),
                (
                    "comparision",
                    models.CharField(
                        choices=[
                            ("exact", "Is Equal To"),
                            ("iexact", "Is Equal To (case insensitive)"),
                            ("contains", "Contains"),
                            ("icontains", "Contains (case insensitive)"),
                            ("in", "Is One Of"),
                            ("gt", "Greater Than"),
                            ("gte", "Greater Than or Equal To"),
                            ("lt", "Less Than"),
                            ("lte", "Less Than or Equal To"),
                            ("startswith", "Starts With"),
                            ("istartswith", "Starts With (case insensitive)"),
                            ("endswith", "Ends With"),
                            ("iendswith", "Ends With (case insensitive)"),
                            ("regex", "Matches Regular Expression"),
                            (
                                "iregex",
                                "Matches Regular Expression (case insensitive)",
                            ),
                            ("isnull", "Is Null"),
                            ("isnotnull", "Is Not Null"),
                            ("true", "Is True"),
                            ("false", "Is False"),
                        ],
                        max_length=20,
                    ),
                ),
                (
                    "param_2",
                    models.CharField(blank=True, max_length=255, null=True),
                ),
                (
                    "signal",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="email_signals.signal",
                    ),
                ),
            ],
        ),
    ]



import typing as _t
import random
import string
from django.db import connection, models
from django.contrib.contenttypes.models import ContentType
from ..models import EmailSignalMixin


def generate_random_string() -> str:
    """Generates a random string."""
    return "".join(
        random.choice(string.ascii_letters)
        for _ in range(random.randint(3, 10))
    )


def generate_random_email() -> str:
    """Generates a random email."""
    return f"{generate_random_string()}@{generate_random_string()}.com"


class TestCustomerModel(models.Model, EmailSignalMixin):
    """A model used to testing purposes. It will imitate a customer model."""

    name = models.CharField(max_length=200, default=generate_random_string)
    email = models.CharField(max_length=200, default=generate_random_email)

    def my_mailing_list(self) -> _t.List[str]:
        return [self.email]

    @classmethod
    def create_table(cls) -> None:
        """Creates the table in the database."""
        if cls._meta.db_table not in connection.introspection.table_names():
            with connection.schema_editor() as schema_editor:
                schema_editor.create_model(cls)
            ContentType.objects.get_or_create(
                app_label="email_signals",
                model="testcustomermodel",
            )

    @classmethod
    def drop_table(cls) -> None:
        """Drops the table from the database."""
        if cls._meta.db_table in connection.introspection.table_names():
            with connection.cursor() as cursor:
                cursor.execute(
                    "DROP TABLE IF EXISTS {};".format(cls._meta.db_table)
                )
            ContentType.objects.filter(
                app_label="email_signals",
                model="testcustomermodel",
            ).delete()

    @classmethod
    def create_record(cls) -> "TestCustomerModel":
        """Creates a record in the database."""
        rec = cls.objects.create()
        rec.save()
        return rec


class TestCustomerOrderModel(models.Model, EmailSignalMixin):
    """A model used to testing purposes. It will imitate a customer order
    model.
    """

    customer = models.ForeignKey(TestCustomerModel, on_delete=models.CASCADE)
    order_number = models.CharField(
        max_length=100, default=generate_random_string, null=True, blank=True
    )

    def my_mailing_list(self) -> _t.List[str]:
        return [self.customer.email]

    @classmethod
    def create_table(cls) -> None:
        """Creates the table in the database."""
        if cls._meta.db_table not in connection.introspection.table_names():
            with connection.schema_editor() as schema_editor:
                schema_editor.create_model(cls)
            ContentType.objects.get_or_create(
                app_label="email_signals",
                model="testcustomerordermodel",
            )

    @classmethod
    def drop_table(cls) -> None:
        """Drops the table from the database."""
        if cls._meta.db_table in connection.introspection.table_names():
            with connection.cursor() as cursor:
                cursor.execute(
                    "DROP TABLE IF EXISTS {};".format(cls._meta.db_table)
                )
            ContentType.objects.filter(
                app_label="email_signals",
                model="testcustomerordermodel",
            ).delete()

    @classmethod
    def create_record(
        cls, customer: TestCustomerModel
    ) -> "TestCustomerOrderModel":
        """Creates a record in the database."""
        rec = cls.objects.create(customer=customer)
        rec.save()
        return rec


class TestM2MModel(models.Model, EmailSignalMixin):
    """A model used to testing purposes. It will imitate a m2m model."""

    customers = models.ManyToManyField(
        TestCustomerModel, related_name="fav_colors"
    )
    fav_colour = models.CharField(
        max_length=50, default=generate_random_string
    )

    @classmethod
    def create_table(cls) -> None:
        """Creates the table in the database."""
        if cls._meta.db_table not in connection.introspection.table_names():
            with connection.schema_editor() as schema_editor:
                schema_editor.create_model(cls)
            ContentType.objects.get_or_create(
                app_label="email_signals",
                model="testM2MModel",
            )

    @classmethod
    def drop_table(cls) -> None:
        """Drops the table from the database."""
        if cls._meta.db_table in connection.introspection.table_names():
            with connection.cursor() as cursor:
                cursor.execute(
                    "DROP TABLE IF EXISTS {};".format(cls._meta.db_table)
                )
            ContentType.objects.filter(
                app_label="email_signals",
                model="testM2MModel",
            ).delete()

    @classmethod
    def create_record(cls, customer: TestCustomerModel) -> "TestM2MModel":
        """Creates a record in the database."""
        rec = cls.objects.create(customer=customer)
        rec.save()
        return rec


class TestOne2OneModel(models.Model, EmailSignalMixin):
    """A model used to testing purposes. It will imitate a one-to-one
    relation.
    """

    customer = models.OneToOneField(
        TestCustomerModel, on_delete=models.CASCADE
    )
    age = models.IntegerField(default=10)

    @classmethod
    def create_table(cls) -> None:
        """Creates the table in the database."""
        if cls._meta.db_table not in connection.introspection.table_names():
            with connection.schema_editor() as schema_editor:
                schema_editor.create_model(cls)
            ContentType.objects.get_or_create(
                app_label="email_signals",
                model="testone2onemodel",
            )

    @classmethod
    def drop_table(cls) -> None:
        """Drops the table from the database."""
        if cls._meta.db_table in connection.introspection.table_names():
            with connection.cursor() as cursor:
                cursor.execute(
                    "DROP TABLE IF EXISTS {};".format(cls._meta.db_table)
                )
            ContentType.objects.filter(
                app_label="email_signals",
                model="testone2onemodel",
            ).delete()

    @classmethod
    def create_record(cls, customer: TestCustomerModel) -> "TestOne2OneModel":
        """Creates a record in the database."""
        rec = cls.objects.create(customer=customer)
        rec.save()
        return rec



from django.urls import path
from . import views

urlpatterns = [
    path(
        "model-attrs/<int:content_type_id>/",
        views.model_attrs,
        name="django_signals_model_attrs",
    ),
]



# Generated by Django 3.2.9 on 2021-12-04 12:58

import ckeditor.fields
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("email_signals", "0004_alter_signal_html_email"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="signal",
            name="html_email",
        ),
        migrations.RemoveField(
            model_name="signal",
            name="plain_text_email",
        ),
        migrations.AddField(
            model_name="signal",
            name="html_message",
            field=ckeditor.fields.RichTextField(
                blank=True, null=True, verbose_name="HTML content"
            ),
        ),
        migrations.AddField(
            model_name="signal",
            name="plain_message",
            field=models.TextField(
                blank=True, null=True, verbose_name="Plain text content"
            ),
        ),
        migrations.AlterField(
            model_name="signal",
            name="content_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="contenttypes.contenttype",
                verbose_name="Model",
            ),
        ),
        migrations.AlterField(
            model_name="signal",
            name="from_email",
            field=models.EmailField(
                blank=True,
                help_text="If not set, `settings.EMAIL_SIGNAL_DEFAULT_SENDER`             with be used.",
                max_length=254,
                null=True,
            ),
        ),
    ]



# Generated by Django 3.2.9 on 2022-09-06 04:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('email_signals', '0006_auto_20211204_1335'),
    ]

    operations = [
        migrations.AlterField(
            model_name='signal',
            name='mailing_list',
            field=models.TextField(help_text='The mailing list to send the signal to. Either enter a comma separated list of emails or the app will search for a function with the same name in the model instance.'),
        ),
    ]



from django.http import HttpRequest, JsonResponse
from django.shortcuts import get_object_or_404
from django.contrib.contenttypes.models import ContentType
from django.contrib.admin.views.decorators import staff_member_required
from . import utils


@staff_member_required
def model_attrs(request: HttpRequest, content_type_id: int) -> JsonResponse:
    """Returns a JSON response with the model attributes for the given content
    type.
    """

    content_type = get_object_or_404(ContentType, pk=content_type_id)
    model = content_type.model_class()
    attrs = utils.get_model_attr_names(model)
    return JsonResponse(attrs, safe=False)



import os
import typing as _t
from pathlib import Path
from django.test import TestCase
from django.conf import settings
from django.db import models
from django.contrib.contenttypes.models import ContentType
from ..models import Signal
from ..registry import add_to_registry
from ..signals import setup as signals_setup


def setup_settings():
    """Helper function to parts of the settings that are used in the tests."""
    settings.TEMPLATES = [
        {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "DIRS": [Path(__file__).resolve().parent.parent.parent],
        }
    ]


class EmailSignalTestCase(TestCase):
    """Wrapper for the Django testcase. Will handle the test models setup
    and teardown.
    """

    @classmethod
    def setUpClass(cls):
        from . import models

        models.TestCustomerModel.create_table()
        models.TestCustomerOrderModel.create_table()
        models.TestM2MModel.create_table()
        models.TestOne2OneModel.create_table()

        cls.Customer = models.TestCustomerModel
        cls.CustomerOrder = models.TestCustomerOrderModel
        cls.M2MModel = models.TestM2MModel
        cls.One2OneModel = models.TestOne2OneModel

        setup_settings()
        super().setUpClass()

    def setUp(self):
        self.customer_rec = self.Customer.create_record()
        self.customer_order_rec = self.CustomerOrder.create_record(
            self.customer_rec
        )

    @classmethod
    def tearDownClass(cls):
        cls.CustomerOrder.drop_table()
        cls.M2MModel.drop_table()
        cls.One2OneModel.drop_table()
        cls.Customer.drop_table()
        super().tearDownClass()

    def tearDown(self):
        self.CustomerOrder.objects.all().delete()
        self.M2MModel.objects.all().delete()
        self.One2OneModel.objects.all().delete()
        self.Customer.objects.all().delete()
        super().tearDown()

    def setup_signals(self):
        """Adds the test models to the signal registry and enables to signals
        to be raised.
        """
        add_to_registry(self.Customer)
        add_to_registry(self.CustomerOrder)
        signals_setup()

    @staticmethod
    def create_signal(
        model_instance: models.Model,
        signal_type: Signal.SignalTypeChoices = Signal.SignalTypeChoices.pre_save,  # noqa E501
        name: str = "Test Signal",
        content_type: _t.Optional[ContentType] = None,
        from_email: str = "test@email.com",
        subject: str = "Test Subject",
        template: _t.Optional[str] = None,
        mailing_list: str = "my_mailing_list",
    ) -> Signal:
        """Create a signal record for a model instance and signal type.

        Args:
            model_instance (models.Model): The model instance to create a
                signal for.
            signal_type (Signal.SignalTypeChoices): The signal type to create
                for the model instance.
            name (str, optional): The name of the signal. Defaults to
                'Test Signal'.
            content_type (ContentType, optional): The content type of the
                model instance. Defaults to the content type of the instance
                provided.
            from_email (str, optional): The from email address to use for
                the signal. Defaults to 'test@email.com'.
            subject (str, optional): The subject of the signal. Defaults to
                'Test Subject'.
            template (str, optional): The template to use for the signal.
        """
        rec = Signal.objects.create(
            name=name,
            content_type=content_type
            or ContentType.objects.get_for_model(model_instance),
            signal_type=signal_type,
            from_email=from_email,
            subject=subject,
            template=template,
            mailing_list=mailing_list,
        )
        rec.save()
        return rec



from django.apps import AppConfig, apps


class EmailSignalsConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "email_signals"
    verbose_name = "Email Signals"

    def ready(self):
        """Find models which should be registered by this app and register
        them.
        """
        from .registry import add_to_registry

        for model in apps.get_models():
            if getattr(model, "EMAIL_SIGNAL_MODEL", False):
                add_to_registry(model)

        # Order matters! We need to register models before we can import
        # `signals` models as it depends on registered models.
        from . import signals

        signals.setup()



from django import template
from django.urls import reverse
from django.urls.exceptions import NoReverseMatch


register = template.Library()


@register.simple_tag
def model_attrs_url():
    """Returns a URL for the model attrs view if it exists."""
    try:
        return reverse("django_signals_model_attrs", args=[1]).replace(
            "1", "<content_type_id>"
        )
    except NoReverseMatch:
        return ""



# Generated by Django 3.2.9 on 2021-12-04 13:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("email_signals", "0005_auto_20211204_1258"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="signal",
            name="to_emails_opt",
        ),
        migrations.AddField(
            model_name="signal",
            name="mailing_list",
            field=models.CharField(
                default=1,
                help_text="The mailing list to send the signal to. Will search for a             function with the same name in the model instance.",
                max_length=100,
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="signal",
            name="template",
            field=models.CharField(
                blank=True,
                help_text="Custom template to use for the email. (Paths relative to             `settings.TEMPLATES[i]['DIRS']`)",
                max_length=100,
                null=True,
            ),
        ),
    ]



from django.test import SimpleTestCase
from .. import constraint_methods


class TestConstraintMethods(SimpleTestCase):
    """Tests the functions in the `constraint_methods` module."""

    def test_exact(self):
        """Tests the `exact` constraint method."""
        self.assertTrue(constraint_methods.exact(1, 1))
        self.assertFalse(constraint_methods.exact(1, 2))

    def test_iexact(self):
        """Tests the `iexact` constraint method."""
        self.assertTrue(constraint_methods.iexact("a", "A"))
        self.assertTrue(constraint_methods.iexact("a", "a"))
        self.assertFalse(constraint_methods.iexact("a", "b"))

    def test_contains(self):
        """Tests the `contains` constraint method."""
        self.assertTrue(constraint_methods.contains("back", "a"))
        self.assertFalse(constraint_methods.contains("a", "b"))

    def test_contains_type_check(self):
        """Tests the `contains` constraint method handles types which are not
        strings.
        """
        self.assertFalse(constraint_methods.contains(1, "a"))
        self.assertFalse(constraint_methods.contains("abc", None))

    def test_icontains(self):
        """Tests the `icontains` constraint method."""
        self.assertTrue(constraint_methods.icontains("BACK", "a"))
        self.assertTrue(constraint_methods.icontains("a", "a"))
        self.assertFalse(constraint_methods.icontains("a", "b"))

    def test_icontains_type_check(self):
        """Tests the `icontains` constraint method handles types which are not
        strings.
        """
        self.assertFalse(constraint_methods.icontains(1, "a"))
        self.assertFalse(constraint_methods.icontains("abc", None))

    def test_gt(self):
        """Tests the `gt` constraint method."""
        self.assertTrue(constraint_methods.gt(2, 1))
        self.assertFalse(constraint_methods.gt(1, 1))
        self.assertFalse(constraint_methods.gt(1, 2))

    def test_gt_type_conversion(self):
        """Tests the `gt` constraint method handles types which are not
        numbers.
        """
        # Should convert to floats and compare successfully.
        self.assertTrue(constraint_methods.gt("1", "0"))
        self.assertFalse(constraint_methods.gt("0", "1"))

        # Should return False if the types cannot be converted.
        self.assertFalse(constraint_methods.gt("a", "b"))

    def test_gte(self):
        """Tests the `gte` constraint method."""
        self.assertTrue(constraint_methods.gte(2, 1))
        self.assertTrue(constraint_methods.gte(1, 1))
        self.assertFalse(constraint_methods.gte(1, 2))

    def test_gte_type_conversion(self):
        """Tests the `gte` constraint method handles types which are not
        numbers.
        """
        # Should convert to floats and compare successfully.
        self.assertTrue(constraint_methods.gte("1", "0"))
        self.assertFalse(constraint_methods.gte("0", "1"))
        self.assertTrue(constraint_methods.gte("1", "1"))

        # Should return False if the types cannot be converted.
        self.assertFalse(constraint_methods.gte("a", "b"))

    def test_lt(self):
        """Tests the `lt` constraint method."""
        self.assertTrue(constraint_methods.lt(1, 2))
        self.assertFalse(constraint_methods.lt(1, 1))
        self.assertFalse(constraint_methods.lt(2, 1))

    def test_lt_type_conversion(self):
        """Tests the `lt` constraint method handles types which are not
        numbers.
        """
        # Should convert to floats and compare successfully.
        self.assertTrue(constraint_methods.lt("0", "1"))
        self.assertFalse(constraint_methods.lt("1", "0"))

        # Should return False if the types cannot be converted.
        self.assertFalse(constraint_methods.lt("a", "b"))

    def test_lte(self):
        """Tests the `lte` constraint method."""
        self.assertTrue(constraint_methods.lte(1, 2))
        self.assertTrue(constraint_methods.lte(1, 1))
        self.assertFalse(constraint_methods.lte(2, 1))

    def test_lte_type_conversion(self):
        """Tests the `lte` constraint method handles types which are not
        numbers.
        """
        # Should convert to floats and compare successfully.
        self.assertTrue(constraint_methods.lte("0", "1"))
        self.assertFalse(constraint_methods.lte("1", "0"))
        self.assertTrue(constraint_methods.lte("1", "1"))

        # Should return False if the types cannot be converted.
        self.assertFalse(constraint_methods.lte("a", "b"))

    def test_startswith(self):
        """Tests the `startswith` constraint method."""
        self.assertTrue(constraint_methods.startswith("abc", "a"))
        self.assertFalse(constraint_methods.startswith("abc", "b"))

    def test_startswith_type_check(self):
        """Tests the `startswith` constraint method handles types which are not
        strings.
        """
        self.assertFalse(constraint_methods.startswith(1, "a"))
        self.assertFalse(constraint_methods.startswith("abc", None))

    def test_istartswith(self):
        """Tests the `istartswith` constraint method."""
        self.assertTrue(constraint_methods.istartswith("ABC", "a"))
        self.assertTrue(constraint_methods.istartswith("abc", "a"))
        self.assertFalse(constraint_methods.istartswith("abc", "b"))

    def test_istartswith_type_check(self):
        """Tests the `istartswith` constraint method handles types which are
        not strings.
        """
        self.assertFalse(constraint_methods.istartswith(1, "a"))
        self.assertFalse(constraint_methods.istartswith("abc", None))

    def test_endswith(self):
        """Tests the `endswith` constraint method."""
        self.assertTrue(constraint_methods.endswith("abc", "c"))
        self.assertFalse(constraint_methods.endswith("abc", "b"))

    def test_endswith_type_check(self):
        """Tests the `endswith` constraint method handles types which are not
        strings.
        """
        self.assertFalse(constraint_methods.endswith(1, "c"))
        self.assertFalse(constraint_methods.endswith("abc", None))

    def test_iendswith(self):
        """Tests the `iendswith` constraint method."""
        self.assertTrue(constraint_methods.iendswith("ABC", "c"))
        self.assertTrue(constraint_methods.iendswith("abc", "c"))
        self.assertFalse(constraint_methods.iendswith("abc", "b"))

    def test_iendswith_type_check(self):
        """Tests the `iendswith` constraint method handles types which are not
        strings.
        """
        self.assertFalse(constraint_methods.iendswith(1, "c"))
        self.assertFalse(constraint_methods.iendswith("abc", None))

    def test_regex(self):
        """Tests the `regex` constraint method."""
        self.assertTrue(constraint_methods.regex("abc", "^a"))
        self.assertFalse(constraint_methods.regex("abc", "^b"))

    def test_iregex(self):
        """Tests the `iregex` constraint method."""
        self.assertTrue(constraint_methods.iregex("Abc1", "^a"))
        self.assertFalse(constraint_methods.iregex("aBc1", "d"))

    def test_isnull(self):
        """Tests the `isnull` constraint method."""
        self.assertTrue(constraint_methods.isnull(None))
        self.assertFalse(constraint_methods.isnull(1))

    def test_isnotnull(self):
        """Tests the `isnotnull` constraint method."""
        self.assertFalse(constraint_methods.isnotnull(None))
        self.assertTrue(constraint_methods.isnotnull(1))

    def test_istrue(self):
        """Tests the `istrue` constraint method."""
        self.assertTrue(constraint_methods.istrue(True))
        self.assertTrue(constraint_methods.istrue("abc"))
        self.assertFalse(constraint_methods.istrue(False))
        self.assertFalse(constraint_methods.istrue(None))
        self.assertFalse(constraint_methods.istrue(""))

    def test_isfalse(self):
        """Tests the `isfalse` constraint method."""
        self.assertTrue(constraint_methods.isfalse(False))
        self.assertTrue(constraint_methods.isfalse(""))
        self.assertFalse(constraint_methods.isfalse(True))
        self.assertFalse(constraint_methods.isfalse("abc"))
        self.assertTrue(constraint_methods.isfalse(None))



# Generated by Django 3.2.9 on 2021-12-04 02:11

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("email_signals", "0002_auto_20211203_1357"),
    ]

    operations = [
        migrations.RenameField(
            model_name="signalconstraint",
            old_name="comparision",
            new_name="comparison",
        ),
        migrations.AlterField(
            model_name="signal",
            name="content_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="contenttypes.contenttype",
                verbose_name="Model (Table)",
            ),
        ),
        migrations.AlterField(
            model_name="signal",
            name="html_email",
            field=models.TextField(
                blank=True, null=True, verbose_name="HTML email"
            ),
        ),
        migrations.AlterField(
            model_name="signal",
            name="to_emails_opt",
            field=models.PositiveIntegerField(
                default=1,
                help_text="The choice of the user to send the email to. For each             integer `i`, the method `get_email_signal_emails_<i>` will be             called on the model instance. The method should return a list of             emails to send to.",
                verbose_name="Mailing list no",
            ),
        ),
        migrations.AlterField(
            model_name="signalconstraint",
            name="param_1",
            field=models.CharField(
                help_text='Will be searched in the instance and signal kwargs recursively. Use "." to show a layer in each attribute.',
                max_length=255,
                verbose_name="Parameter 1",
            ),
        ),
        migrations.AlterField(
            model_name="signalconstraint",
            name="param_2",
            field=models.CharField(
                blank=True,
                help_text='Will be searched in the instance and signal kwargs recursively. Use "." to show a layer in each attribute. Also supports primitive values.',
                max_length=255,
                null=True,
                verbose_name="Parameter 2",
            ),
        ),
    ]



from types import SimpleNamespace
from django.test import SimpleTestCase
from .. import utils
from .testcase import EmailSignalTestCase


class TestConvertToPrimitive(SimpleTestCase):
    """Unittests for the `convert_to_primitive` utility function."""

    def test_convert_to_true(self):
        """Test the `convert_to_primitive` method where the paramater provided
        is 'true'
        """
        self.assertEqual(utils.convert_to_primitive("true"), True)
        self.assertEqual(utils.convert_to_primitive("tRuE"), True)

    def test_convert_to_false(self):
        """Test the `convert_to_primitive` method where the paramater provided
        is 'false'
        """
        self.assertEqual(utils.convert_to_primitive("false"), False)
        self.assertEqual(utils.convert_to_primitive("fAlSe"), False)

    def test_convert_to_none(self):
        """Test the `convert_to_primitive` method where the paramater provided
        is 'none'
        """
        self.assertEqual(utils.convert_to_primitive("none"), None)
        self.assertEqual(utils.convert_to_primitive("nOnE"), None)
        self.assertEqual(utils.convert_to_primitive("null"), None)
        self.assertEqual(utils.convert_to_primitive("nuLL"), None)

    def test_convert_to_float(self):
        """Test the `convert_to_primitive` method where the paramater provided
        is a float
        """
        self.assertEqual(utils.convert_to_primitive("1.0"), 1.0)
        self.assertEqual(utils.convert_to_primitive("1.1"), 1.1)

    def test_convert_to_bad_float(self):
        """Test the `convert_to_primitive` method where the paramater provided
        is not a flat but has a '.' in it. It should return the value back as
        a string.
        """
        self.assertEqual(utils.convert_to_primitive("1.1.1"), "1.1.1")

    def test_convert_to_int(self):
        """Test the `convert_to_primitive` method where the paramater provided
        is an int
        """
        self.assertEqual(utils.convert_to_primitive("1"), 1)
        self.assertEqual(utils.convert_to_primitive("2"), 2)

    def test_failed_convert_to_primitive(self):
        """Test the `convert_to_primitive` method where the paramater provided
        is not a valid value.
        """
        self.assertEqual(utils.convert_to_primitive("a"), "a")


class TestGetParamFromObj(EmailSignalTestCase):
    """Unittests for the `get_param_from_obj` utility function."""

    def test_dict(self):
        """Test the function returns the current value from a `dict`."""
        self.assertEqual(utils.get_param_from_obj("a", {"a": 1}), (True, 1))

    def test_multi_layered_dict(self):
        """Test the function returns the current value from a `dict`."""
        self.assertEqual(
            utils.get_param_from_obj("a.b.c", {"a": {"b": {"c": 1}}}),
            (True, 1),
        )

    def test_iterable(self):
        """Test the function returns the current value from a iterable."""
        self.assertEqual(utils.get_param_from_obj("0", [1, 2]), (True, 1))
        self.assertEqual(utils.get_param_from_obj("0", (1, 2)), (True, 1))

    def test_multi_layered_iterable(self):
        """Test the function returns the current value from a iterable."""
        self.assertEqual(
            utils.get_param_from_obj("0.1", [[1, 2, 3], [4, 5, 6], [7, 8, 9]]),
            (True, 2),
        )
        self.assertEqual(
            utils.get_param_from_obj("0.1", ((1, 2, 3), (4, 5, 6), (7, 8, 9))),
            (True, 2),
        )

    def test_simplenamespace(self):
        """Test the function returns the current value from a `SimpleNamespace`
        object.
        """
        self.assertEqual(
            utils.get_param_from_obj("a", SimpleNamespace(a=1)),
            (True, 1),
        )

    def test_model_instance(self):
        """Test that the function is able to travel into a model instance."""

        self.Customer.objects.create(name="Util Test Customer").save()

        self.assertEqual(
            utils.get_param_from_obj(
                "0.name",
                self.Customer.objects.filter(name="Util Test Customer"),
            ),
            (True, "Util Test Customer"),
        )


class TestGetModelAttrNames(EmailSignalTestCase):
    """Unittests for the `get_model_attr_names` utility function."""

    def test_get_model_attr_names_customer(self):
        """Test the `get_model_attr_names` function contains the correct
        attributes when provided the customer model."""
        model_attr_names = utils.get_model_attr_names(self.Customer)

        for field in ("id", "name", "email"):
            self.assertIn(field, model_attr_names)

    def test_get_model_attr_names_customer_order(self):
        """Test the `get_model_attr_names` function contains the correct
        attributes when provided the customer order model."""

        model_attr_names = utils.get_model_attr_names(self.CustomerOrder)

        for field in ("customer", "order_number"):
            self.assertIn(field, model_attr_names)

        for field in ("name", "email"):
            self.assertIn(field, model_attr_names["customer"])

    def test_get_model_attrs_m2m_model(self):
        """Test the `get_model_attr_names` function contains the correct
        attributes when provided the test M2M model."""

        model_attr_names = utils.get_model_attr_names(self.M2MModel)

        for field in ("id", "customers", "fav_colour"):
            self.assertIn(field, model_attr_names)

    def test_get_model_attrs_one2one_model(self):
        """Test the `get_model_attr_names` function contains the correct
        attributes when provided the test one2one model."""

        model_attr_names = utils.get_model_attr_names(self.One2OneModel)

        for field in ("id", "customer", "age"):
            self.assertIn(field, model_attr_names)

        for field in ("id", "name", "email"):
            self.assertIn(field, model_attr_names["customer"])


class TestAddContextToString(EmailSignalTestCase):
    """Unittests for the `add_context_to_string` utility function."""

    def test_with_nested_dict(self):
        """Test the `add_context_to_string` function works with a nested
        dict.
        """
        self.assertEqual(
            utils.add_context_to_string("test {{ a.a }}", {"a": {"a": 1}}),
            "test 1",
        )

    def test_with_nested_list(self):
        """Test the `add_context_to_string` function works with a list."""
        self.assertEqual(
            utils.add_context_to_string("test {{ a.1 }}", {"a": [1, 2, 3]}),
            "test 2",
        )

    def test_with_model(self):
        """Test the `add_context_to_string` function works with a model."""

        self.assertEqual(
            utils.add_context_to_string(
                "test {{ order.customer.name }}",
                {
                    "order": self.customer_order_rec,
                },
            ),
            f"test {self.customer_rec.name}",
        )

    def test_with_missing_context(self):
        """Test that the `add_context_to_string` function replaces missing
        context with an empty string.
        """
        self.assertEqual(
            utils.add_context_to_string("test {{ a.a }}", {}),
            "test ",
        )



from django.test import TestCase
from ..models import Signal
from .. import registry


class TestRegistry(TestCase):
    """Tests the functions in the registry module."""

    def setUp(self):
        # Flush the registry
        registry.registered_models = {}

    def test_model_str(self):
        """Tests the model_str function."""
        self.assertEqual(registry.model_str(Signal), "email_signals.signal")

    def test_add_to_registry(self):
        """Tests the add_to_registry function."""
        registry.add_to_registry(Signal)
        self.assertEqual(list(registry.registered_models.values()), [Signal])

    def test_model_in_registry(self):
        """Tests the model_in_registry function."""
        self.assertFalse(registry.model_in_registry(Signal))
        registry.add_to_registry(Signal)
        self.assertTrue(registry.model_in_registry(Signal))

    def test_registered_content_types(self):
        """Tests the registered_content_types function. It is expected to
        return a queryset of content types for the registered models.
        """
        registry.add_to_registry(Signal)
        results = registry.registered_content_types()
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].model_class(), Signal)
